/********************************************************************************
 * Copyright (c) 2023 ssm.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/
import { EdgeReceiver } from "./EdgeReceiver";
import { EdgeType } from "./EdgeType";
import { NamedElement } from "./NamedElement";
import { SimpleNode } from "./SimpleNode";
import { TransformationEdge } from "./TransformationEdge";
import { TransformationConfiguration } from "./TransformationConfiguration";
import { TransformationProcessor } from "./TransformationProcessor";
import { KicoCloneable } from "@pragmatic-programming/kico";
export type IHNode = SimpleNode | IHGraph;
/**
 * The IHGraph class represents a graph that can be used to model induced hierarchies.
 * It is a directed, labeled graph that contains nodes and edges.
 * Besides the usual graph operations it contains specialized methods to manipulate graph cliques, determined by their edge types, and to create and resolve induced hierarchies.
 * The data structure extends the NamedElement class and is hence uniquely identifiable.
 * It implements the EdgeReceiver interface to receive edges from other nodes/graphs and KiCoCloneable for the KiCo environments to be used in the dynamic compiler.
 */
export declare class IHGraph extends NamedElement implements EdgeReceiver, KicoCloneable {
    /****************************************
     *
     * Fields & Constructor
     *
     */
    protected parent: IHGraph | undefined;
    protected nodes: IHNode[];
    protected edgeTypes: EdgeType[];
    protected incomingEdges: TransformationEdge[];
    protected outgoingEdges: TransformationEdge[];
    protected transformationConfiguration: TransformationConfiguration;
    /**
     * Constructor of the IHGraph class.
     * @param parent The parent graph of this graph. If undefined, this graph is the root graph.
     */
    constructor(parent?: IHGraph | undefined);
    /**
     * Tells KiCo that the structure is mutable.
     */
    isMutable(): boolean;
    /**
     * Retrieve the parent of this graph.
     * @returns The parent graph of this graph. If undefined, this graph is the root graph.
     */
    getParent(): IHGraph | undefined;
    /**
     * Set a new parent graph.
     * @param parent The parent.
     */
    setParent(parent: IHGraph): void;
    /****************************************
     *
     * Create
     *
     */
    /**
     * Creates a new simple node and adds it to the graph structure.
     * @param id The name of the source node.
     * @returns The source node.
     */
    createSimpleNode(id: string): SimpleNode;
    /**
     * Creates a new transformation edge between two IHNodes.
     * @param type The used edge type.
     * @param sourceNode The source node of the edge.
     * @param targetNode The target node of the edge.
     * @returns The new edge.
     */
    createTransformationEdge(type: EdgeType, sourceNode: IHNode, targetNode: IHNode): TransformationEdge;
    /**
     * Creates a new edge type.
     * @param id The id of the edge type.
     * @param priority The priority of the edge type.
     * @returns The edge type.
     */
    createEdgeType(id: string, priority: number): EdgeType;
    /****************************************
     *
     * Getter / Setter for nodes
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Retrieves all nodes of the graph.
     * @returns A list of nodes of the graph.
     */
    getNodes(): IHNode[];
    /**
     * Retrieves all node ids of the graph.
     * @returns A list of node ids of the graph.
     */
    getNodeIds(): string[];
    /**
     * Retrieves all simple nodes of the graph. Simple nodes do not contain hierarchy.
     * @returns A list of simple nodes of the graph.
     */
    getSimpleNodes(): SimpleNode[];
    /**
     * Retrieves all shallow graph nodes, which contain hierarchy, of the graph.
     * @returns A list of graph nodes of the graph.
     */
    getGraphNodes(): IHGraph[];
    /**
     * Retrieves all nodes, also included in hierarchy, of the graph.
     * @returns A list of graph nodes of the graph.
     */
    getDeepNodes(): IHNode[];
    /**
     * Retrieve all source nodes of the graph even if included in hierarchy.
     * @returns A list source node of the graph.
     */
    getDeepSimpleNodes(): IHNode[];
    /**
     * Retrieve all source nodes of the graph. Source nodes to not have incoming edges.
     * @returns A list of source nodes without incoming edges.
     */
    getSourceNodes2(): SimpleNode[];
    /**
     * Retrieve all sink nodes of the graph. Sink nodes to not have outgoing edges.
     * @returns A list of sink nodes without outgoing edges.
     */
    getSinkNodes(): SimpleNode[];
    /**
     * Retrieves the node instance of the node with the provided id. Also works with hierarchy.
     * @param id The id of the node.
     * @returns The queried node instance or null if the node does not exist.
     */
    getNodeById(id: string): IHNode | undefined;
    /**
     * Retrieves the source node of the edge with the highest edge priority. Works shallowly and does not look into hierarchies.
     * @returns The source node of the edge with the highest edge priority.
     */
    getHighestPriorityShallowNode(): IHNode;
    /**
     * Adds a node to the graph. It automatically sets the parent of the node to this graph.
     * @param node The node that should be added to the graph.
     */
    addNode(node: IHNode): void;
    /**
     * Removes the given node from the graph.
     * @param node The node that should be removed from the graph.
     */
    removeNode(node: IHNode): void;
    /**
     * Removes the node with the given id from the graph. Throws an error if the node is not included in the graph.
     * @param id The id of the node that should be removed from the graph.
     */
    removeNodeById(id: string): void;
    /****************************************
     *
     * Getter / Setter for edges of the EdgeReceiver interface
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Returns the incoming edges connected to this node. Implements part of the EdgeReceiver interface.
     * @returns A list of incoming edges.
     */
    getIncomingEdges(): TransformationEdge[];
    /**
     * Returns the outgoing edges connected to this node. Implements part of the EdgeReceiver interface.
     * @returns A list of outgoing edges.
     */
    getOutgoingEdges(): TransformationEdge[];
    /**
     * Adds an incoming edge to the node. Implements part of the EdgeReceiver interface.
     * @param edge The incoming edge.
     */
    addIncomingEdge(edge: TransformationEdge): void;
    /**
     * Adds an outgoing edge to the node. Implements part of the EdgeReceiver interface.
     * @param edge The outgoing edge.
     */
    addOutgoingEdge(edge: TransformationEdge): void;
    /**
     * Removes an incoming edge from the node. Implements part of the EdgeReceiver interface.
     * @param edge The incoming edge.
     */
    removeIncomingEdge(edge: TransformationEdge): void;
    /**
     * Removes an outgoing edge from the node. Implements part of the EdgeReceiver interface.
     * @param edge The outgoing edge.
     */
    removeOutgoingEdge(edge: TransformationEdge): void;
    /****************************************
     *
     * Getter / Setter for edges of the graph
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Retrieves all edges of the graph.
     * @returns A list of transformation edges of the graph.
     */
    getEdges(): TransformationEdge[];
    /**
     * Retrieves all simple node edges of the graph.
     * @returns A list of transformation edges of the graph only considering simple nodes.
     */
    getSimpleNodeEdges(): TransformationEdge[];
    /**
     * Retrieves all edges of the graph including edges from nested hierarchies.
     * @returns A list of transformation edges of the graph including edges from nested hierarchies.
     */
    getDeepEdges(): TransformationEdge[];
    /**
     * Removes a transformation edge from the graph.
     * @param edge The edge that should be removed from the graph.
     */
    removeEdge(edge: TransformationEdge): void;
    /**
     * Removes all edges in the graph that correspond to the source and target node ids that match the provided edge.
     * @param edge The edge that is used to identify source and target nodes with the same ids.
     */
    removeEdgeBySourceTargetIds(edge: TransformationEdge): void;
    /****************************************
     *
     * Getter / Setter for edge types
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Retrieves the edge type of the graph.
     * @returns A list of edge types.
     */
    getEdgeTypes(): EdgeType[];
    /**
     * Retrieves the edge type ids of the graph.
     * @returns A list of edge type ids.
     */
    getEdgeTypeIds(): string[];
    /**
     * Retrieves the edge type of the graph.
     * @param id The id of the edge type.
     * @returns The edge type.
     */
    getEdgeTypeById(id: string): EdgeType | undefined;
    /**
     * Checks whether or not a specific edge type is included in the graph.
     * Here, the check returns true if the id, the priority, and the immediate state a equal.
     * It is not mandatory to include the identical object.
     * @param edgeType The edge type that should be checked.
     * @returns True if an identical edge type is present in the graph.
     */
    hasEdgeType(edgeType: EdgeType): boolean;
    /**
     * Retrieves the edge type of the graph that corresponds to a given edge type determined by the id, the priority, and the immediate state.
     */
    getShallowEdgeType(edgeType: EdgeType): EdgeType;
    /**
     * Adds an edge type to the graph.
     * @param edgeType The edge type that should be added to the graph.
     */
    addEdgeType(edgeType: EdgeType): void;
    /**
     * Retrieves the highest priority of the shallow edges of the graph.
     * Throws an error if there are no edges in the graph.
     * @returns The highest priority of the shallow edges of the graph.
     */
    getHighestShallowPriority(): number;
    /**
     * Retrieves the highest priority of the deep edges of the graph.
     * Throws an error if there are no edges in the graph.
     * @returns The highest priority of the deep edges of the graph.
     */
    getHighestDeepPriority(): number;
    /**
     * Removes specific edge type from the graph.
     * @param edgeType The edge type that should be removed from the graph.
     */
    removeEdgeType(edgeType: EdgeType): void;
    /**
     * Removes unused edges types from and adds missing edge types to the graph.
     */
    revalidateEdgeTypes(): void;
    /****************************************
     *
     * Getter / Setter for cliques
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Retrieves the clique, determined by a given edge type, that includes the provided node.
     * The clique is itself an IHGraph and clones the nodes and edges of the original graph.
     * Implemented as breadth-first search.
     * @returns The clique as IHGraph.
     */
    getClique(node: IHNode, edgeType: EdgeType): IHGraph;
    /**
     * Retrieves the next clique determined by the highest priority of the edges and edge types.
     * @returns The next clique determined by the highest priority.
     */
    getNextClique(): IHGraph;
    /**
     * Retrieves all immediate cliques.
     * @returns A list of immediate cliques.
     */
    getImmediateCliques(): IHGraph[];
    /**
     * Adds a clique to the graph. All nodes and edges are added to the graph. The enclosing node is ignored.
     * The nodes are added as they are and not cloned beforehand.
     * @param clique The clique that contains the nodes and edges that should be added to the graph.
     */
    addClique(clique: IHGraph): void;
    /**
     * Replaces a selected clique in the graph with a different clique.
     * Therefore, re-route all edges from outside the clique to the new one.
     * If the new clique only contains one node, all edges will be re-routed to the node.
     * If there exist a node with the same id as the original node, this node will be the target.
     * Otherwise, they will lead to the first node.
     * @param clique The selected clique in the graph.
     * @param replacement The new replacement clique.
     * @param addCliqueNodes Determines if the replacement nodes should be added by the function, which is the default. If set to false, the function only returns the edges that would be re-routed for post-processing.
     * @returns The mapping of the re-routed edges. The first list contains the edges with source nodes outside of the clique meaning the target node lies within the new clique and vice versa in the second list.
     */
    replaceClique(clique: IHGraph, replacement: IHGraph, addCliqueNodes?: boolean): [TransformationEdge[], TransformationEdge[]];
    /**
     * Removes all nodes and edges in the clique from the graph. The nodes and edges are identified by id and not via instances.
     * @param clique The clique containing the nodes and edges that should be removed from the graph.
     */
    removeClique(clique: IHGraph): void;
    /****************************************
     *
     * Induced Hierarchies
     *
     */
    /**
     * Creates a new graph with induced hierarchy. Therefore, hierarchical graph nodes are introduced for each unique clique.
     * Edges of nodes that are moved inside the hierarchy will be re-routed to the new graph node.
     * @param maxIterations The maximum number of iterations. Default is 100.
     * @param iteration The current iteration of the induced hierarchy. Used for recursion. Default is 0.
     * @returns A new graph with induced hierarchy.
     */
    getInducedHierarchy(maxIterations?: number, iteration?: number): IHGraph;
    /**
     * Resolves a hierarchical graph by flattening the hierarchy.
     * The edges of the graph nodes are re-routed to nodes inside the clique. If there are original node annotations, these will be used. Otherwise, edges will be re-routed to the first node of the graph node, which is semantically sound.
     * @returns The flattened graph.
     */
    getFlattenedHierarchy(): IHGraph;
    /****************************************
     *
     * Getter / Setter for transformation configurations
     * The order is generic getter, specialized getters, add, update, remove.
     *
     */
    /**
     * Retrieves the transformation configuration of the graph.
     * @returns The transformation configuration of the graph.
     */
    getTransformationConfiguration(): TransformationConfiguration;
    /**
     * Sets the transformation for a single edge type.
     * @param edgeType The edge type in question.
     * @param processor The Kico transformation processor that should handle this edge type.
     */
    setEdgeTypeTransformation(edgeType: EdgeType, processor: typeof TransformationProcessor): void;
    /**
     * Sets the transformation for a single edge type identified by its id.
     * @param edgeTypeId The edge type id of the edge type in question.
     * @param processor The Kico transformation processor that should handle this edge type.
     */
    setEdgeTypeTransformationById(edgeTypeId: string, processor: typeof TransformationProcessor): void;
    /****************************************
     *
     * Equals & Clone
     *
     */
    /**
     * Checks whether or not the graph is equal to another graph.
     * A graph is equal to another graph or clique if all nodes (ids), edge types, and edges are the same.
     * The order of the nodes and edges does not matter.
     * @param other The graph that should be compared to this graph.
     * @returns True if the graphs are equal.
     * @override
     */
    equals(other: NamedElement): boolean;
    /**
     * Clones the IHGraph with element order preserved. If provided, only including nodes and edge types that are provided.
     * An empty array can be used to lift all restrictions (since cloning an empty graph is not really useful and increases the usability for hierarchical graphs).
     *
     * @returns A clone of the graph.
     */
    clone(nodes?: IHNode[], types?: EdgeType[]): IHGraph;
    /**
     * Clones the IHGraph with element order preserved. If provided, only including nodes and edge types that are provided.
     * An empty array can be used to lift all restrictions (since cloning an empty graph is not really useful and increases the usability for hierarchical graphs).
     *
     * @returns A **clone** of the graph,
     * a **node mapping** from the original nodes to the cloned nodes,
     * a **type mapping** from the original edge types to the cloned edge types,
     * and an **edge mapping** from the original edges to the cloned edges.
     */
    cloneWithMappings(nodes?: IHNode[], types?: EdgeType[]): [IHGraph, Map<IHNode, IHNode>, Map<EdgeType, EdgeType>, Map<TransformationEdge, TransformationEdge>];
    clear(): IHGraph;
    /****************************************
     *
     * Serialize & Stringify
     *
     */
    /**
     * Serializes the graph to a JSON string.
     * @param includeAnnotations Determines if the annotations should be included in the serialization. Default is false.
     * @returns The serialized graph as JSON string.
     */
    serialize(includeAnnotations?: boolean): string;
    /**
     * Creates a string representation of the graph for debug purposes. The string contains nodes, edges, and edge type with unique object identifiers.
     * @param indent Indentation of the string. Default is "".
     * @returns A debug string representation of the graph.
     */
    toStringDebugGraph(indent?: string): string;
    /**
     * Consistency check for the graph to be used in assertions.
     * Mainly used in internal operations and debugging.
     * @returns true if everything is consistent.
     */
    consistency(): boolean;
}
